@@ -1,12 +1,16 @@
 package Treex::Tool::Parser::MSTperl::FeaturesControl;
 {
-  $Treex::Tool::Parser::MSTperl::FeaturesControl::VERSION = '0.08268';
+  $Treex::Tool::Parser::MSTperl::FeaturesControl::VERSION = '0.11306';
 }
 
 use Moose;
 use autodie;
 use Carp;
 
+use Treex::Tool::Parser::MSTperl::ModelAdditional;
+
+# TODO dynamic features
+
 has 'config' => (
     isa      => 'Treex::Tool::Parser::MSTperl::Config',
     is       => 'ro',
@@ -14,7 +18,7 @@ has 'config' => (
     weak_ref => '1',
 );
 
-# FEATURES for unlabelled parsing
+# FEATURES
 
 has 'feature_count' => (
     is  => 'rw',
@@ -60,7 +64,14 @@ has 'array_features' => (
     default => sub { {} },
 );
 
-# SIMPLE FEATURES for unlabelled parser
+# features containing dynamic simple features
+has 'dynamic_features' => (
+    is      => 'rw',
+    isa     => 'HashRef[Int]',
+    default => sub { {} },
+);
+
+# SIMPLE FEATURES
 
 has 'simple_feature_count' => (
     is  => 'rw',
@@ -104,6 +115,15 @@ has 'array_simple_features' => (
     default => sub { {} },
 );
 
+# simple features that must be always recomputed
+# because their value cannot be always computed from input data
+# (for labeller - parent's label, brother's label etc.)
+has 'dynamic_simple_features' => (
+    is      => 'rw',
+    isa     => 'HashRef[Int]',
+    default => sub { {} },
+);
+
 # # simple features that get more than 1 argument as input
 # has 'multiarg_simple_features' => (
 #     is => 'rw',
@@ -129,6 +149,18 @@ has 'edge_features_cache' => (
     default => sub { {} },
 );
 
+has pmi_model => (
+    is      => 'rw',
+    isa     => 'Maybe[Treex::Tool::Parser::MSTperl::ModelAdditional]',
+    default => undef,
+);
+
+has cprob_model => (
+    is      => 'rw',
+    isa     => 'Maybe[Treex::Tool::Parser::MSTperl::ModelAdditional]',
+    default => undef,
+);
+
 sub BUILD {
     my ($self) = @_;
 
@@ -157,7 +189,8 @@ sub set_feature {
     } else {
 
         # get simple features
-        my $isArrayFeature = 0;
+        my $isArrayFeature   = 0;
+        my $isDynamicFeature = 0;
         my @simple_features_indexes;
         my %simple_features_hash;
         foreach my $simple_feature_code ( split( /\|/, $feature_code ) ) {
@@ -182,6 +215,9 @@ sub set_feature {
             if ( $self->array_simple_features->{$simple_feature_index} ) {
                 $isArrayFeature = 1;
             }
+            if ( $self->dynamic_simple_features->{$simple_feature_index} ) {
+                $isDynamicFeature = 1;
+            }
             push @simple_features_indexes, $simple_feature_index;
         }
 
@@ -195,6 +231,9 @@ sub set_feature {
         if ($isArrayFeature) {
             $self->array_features->{$feature_index} = 1;
         }
+        if ($isDynamicFeature) {
+            $self->dynamic_features->{$feature_index} = 1;
+        }
     }
 
     return;
@@ -228,8 +267,9 @@ sub set_simple_feature {
                 ") for parent node.\n";
         }
 
-        # first/second node feature
-    } elsif ( $simple_feature_code =~ /^([12])\.([a-z0-9_]+)$/ ) {
+        # first/second/(left sibling)/(right sibling)/Grandparent/grandchildren
+        # node feature
+    } elsif ( $simple_feature_code =~ /^([12gGlr])\.([a-z0-9_]+)$/ ) {
 
         $simple_feature_field = $2;
 
@@ -241,6 +281,22 @@ sub set_simple_feature {
 
             # second node feature
             $simple_feature_sub = \&{feature_second};
+        } elsif ( $1 eq 'g' ) {
+
+            # grandchildren node feature
+            $simple_feature_sub = \&{feature_grandchildren};
+        } elsif ( $1 eq 'G' ) {
+
+            # grandparent node feature
+            $simple_feature_sub = \&{feature_grandparent};
+        } elsif ( $1 eq 'l' ) {
+
+            # left sibling edge child feature
+            $simple_feature_sub = \&{feature_left_sibling};
+        } elsif ( $1 eq 'r' ) {
+
+            # right sibling edge child feature
+            $simple_feature_sub = \&{feature_right_sibling};
         } else {
             croak "Assertion failed!";
         }
@@ -248,19 +304,34 @@ sub set_simple_feature {
         # function feature
     } elsif (
         $simple_feature_code
-        =~ /^([12\.a-z]+|[A-Z]+)\([-a-z0-9_,]*\)$/
+        =~ /^([12gGlr\.a-z]+|[A-Z]+)\([-a-z0-9_,]*\)$/
         )
     {
         my $function_name = $1;
         $simple_feature_sub =
             $self->get_simple_feature_sub_reference($function_name);
 
-        if ( $function_name eq 'between' || $function_name eq 'foreach' ) {
+        if ($function_name eq 'between'
+            || $function_name eq 'foreach'
+            || substr( $function_name, 0, 2 ) eq 'g.'
+            )
+        {
 
             # array function
             $self->array_simple_features->{$simple_feature_index} = 1;
         }
 
+        if ($function_name eq 'LABEL'
+            || $function_name eq 'l.label' || $function_name eq 'prevlabel'
+            || $function_name eq 'G.label'
+            || $function_name eq 'g.label'
+            )
+        {
+
+            # dynamic feature
+            $self->dynamic_simple_features->{$simple_feature_index} = 1;
+        }
+
         # set $simple_feature_field
         if ( $simple_feature_code =~ /$function_name\(\)$/ ) {
 
@@ -319,7 +390,10 @@ sub set_simple_feature {
 # is not present)
 # TODO maybe not returning a value is still a valuable information -> include?
 sub get_all_features {
-    my ( $self, $edge ) = @_;
+
+    # Edge; 0: all features, 1: only dynamic, -1: only non-dynamic
+    # either get only dynamic features or get all but dynamic features
+    my ( $self, $edge, $only_dynamic_features ) = @_;
 
     # try to get features from cache
     # TODO: cache not used now and probably does not even work:
@@ -344,19 +418,34 @@ sub get_all_features {
         $feature_index++
         )
     {
-        my $feature_value =
-            $self->get_feature_value( $feature_index, $simple_feature_values );
-        if ( $self->array_features->{$feature_index} ) {
-
-            #it is an array feature, the returned value is an array reference
-            foreach my $value ( @{$feature_value} ) {
-                push @features, "$feature_index:$value";
-            }
+        if ($only_dynamic_features
+            && $only_dynamic_features == 1
+            && !$self->dynamic_features->{$feature_index}
+            )
+        {
+            next;
+        } elsif (
+            $only_dynamic_features
+            && $only_dynamic_features == -1
+            && $self->dynamic_features->{$feature_index}
+            )
+        {
+            next;
         } else {
+            my $feature_value =
+                $self->get_feature_value( $feature_index, $simple_feature_values );
+            if ( $self->array_features->{$feature_index} ) {
 
-            #it is not an array feature, the returned value is a string
-            if ( $feature_value ne '' ) {
-                push @features, "$feature_index:$feature_value";
+                #it is an array feature, the returned value is an array reference
+                foreach my $value ( @{$feature_value} ) {
+                    push @features, "$feature_index:$value";
+                }
+            } else {
+
+                #it is not an array feature, the returned value is a string
+                if ( $feature_value ne '' ) {
+                    push @features, "$feature_index:$feature_value";
+                }
             }
         }
     }
@@ -476,8 +565,14 @@ sub get_simple_feature_values_array {
 
 my %simple_feature_sub_references = (
     'LABEL'             => \&{feature_parent_label},
+    'prevlabel'         => \&{feature_previous_label},
+    'l.label'           => \&{feature_previous_label},
+    'G.label'           => \&{feature_grandparent_label},
+    'g.label'           => \&{feature_grandchildren_label},
     'distance'          => \&{feature_distance},
+    'G.distance'        => \&{feature_grandparent_distance},
     'attdir'            => \&{feature_attachement_direction},
+    'G.attdir'          => \&{feature_grandparent_attachement_direction},    # grandparent to child
     'preceding'         => \&{feature_preceding_child},
     'PRECEDING'         => \&{feature_preceding_parent},
     '1.preceding'       => \&{feature_preceding_first},
@@ -506,6 +601,19 @@ my %simple_feature_sub_references = (
     'CHILDNO'           => \&{feature_number_of_parents_children},
     'substr'            => \&{feature_substr_child},
     'SUBSTR'            => \&{feature_substr_parent},
+    'pmi'               => \&{feature_pmi},
+    'pmibucketed'       => \&{feature_pmi_bucketed},
+    'pmirounded'        => \&{feature_pmi_rounded},
+    'pmid'              => \&{feature_pmi_d},
+    'cprob'             => \&{feature_cprob},
+    'cprobbucketed'     => \&{feature_cprob_bucketed},
+    'cprobrounded'      => \&{feature_cprob_rounded},
+
+    # obsolete
+    #    'pmitworounded'     => \&{feature_pmi_2_rounded},
+    #    'pmithreerounded'   => \&{feature_pmi_3_rounded},
+    #    'cprobtworounded'   => \&{feature_cprob_2_rounded},
+    #    'cprobthreerounded' => \&{feature_cprob_3_rounded},
 );
 
 sub get_simple_feature_sub_reference {
@@ -518,13 +626,37 @@ sub get_simple_feature_sub_reference {
     }
 }
 
+# returns undef if there is no grandparent, i.e. the parent is the root
+sub get_grandparent {
+    my ( $self, $edge ) = @_;
+
+    return ( $edge->parent )->parent;
+}
+
 sub feature_distance {
     my ( $self, $edge ) = @_;
 
-    my $distance = $edge->parent->ord - $edge->child->ord;
+    return $self->feature_distance_generic( $edge->parent, $edge->child );
+}
+
+sub feature_grandparent_distance {
+    my ( $self, $edge ) = @_;
+
+    my $grandparent = $self->get_grandparent($edge);
+    if ( defined $grandparent ) {
+        return $self->feature_distance_generic( $edge->parent, $edge->child );
+    } else {
+        return '#novalue#';
+    }
+}
+
+sub feature_distance_generic {
+    my ( $self, $node1, $node2 ) = @_;
+
+    my $distance = $node1->ord - $node2->ord;
 
     my $bucket = $self->config->distance2bucket->{$distance};
-    if ($bucket) {
+    if ( defined $bucket ) {
         return $bucket;
     } else {
         if ( $distance <= $self->config->minBucket ) {
@@ -538,7 +670,28 @@ sub feature_distance {
 sub feature_attachement_direction {
     my ( $self, $edge ) = @_;
 
-    if ( $edge->parent->ord < $edge->child->ord ) {
+    return $self->feature_attachement_direction_generic(
+        $edge->parent, $edge->child
+    );
+}
+
+sub feature_grandparent_attachement_direction {
+    my ( $self, $edge ) = @_;
+
+    my $grandparent = $self->get_grandparent($edge);
+    if ( defined $grandparent ) {
+        return $self->feature_attachement_direction_generic(
+            $edge->parent, $edge->child
+        );
+    } else {
+        return '#novalue#';
+    }
+}
+
+sub feature_attachement_direction_generic {
+    my ( $self, $node1, $node2 ) = @_;
+
+    if ( $node1->ord < $node2->ord ) {
         return -1;
     } else {
         return 1;
@@ -555,11 +708,44 @@ sub feature_parent {
     return ( $edge->parent->fields->[$field_index] );
 }
 
+sub feature_grandparent {
+    my ( $self, $edge, $field_index ) = @_;
+
+    my $grandparent = $self->get_grandparent($edge);
+    if ( defined $grandparent ) {
+        return ( $grandparent->fields->[$field_index] );
+    } else {
+        return '#novalue#';
+    }
+}
+
 sub feature_parent_label {
     my ( $self, $edge ) = @_;
     return ( $edge->parent->label );
 }
 
+sub feature_previous_label {
+    my ( $self, $edge ) = @_;
+
+    my $left_sibling = $self->get_left_sibling($edge);
+    if ( defined $left_sibling ) {
+        return ( $left_sibling->child->label );
+    } else {
+        return $self->config->SEQUENCE_BOUNDARY_LABEL;
+    }
+}
+
+sub feature_grandparent_label {
+    my ( $self, $edge ) = @_;
+
+    my $grandparent = $self->get_grandparent($edge);
+    if ( defined $grandparent ) {
+        return ( $grandparent->label );
+    } else {
+        return '#novalue#';
+    }
+}
+
 sub feature_first {
     my ( $self, $edge, $field_index ) = @_;
     return ( $edge->first->fields->[$field_index] );
@@ -573,12 +759,34 @@ sub feature_second {
 sub feature_left_sibling {
     my ( $self, $edge, $field_index ) = @_;
 
+    my $left_sibling = $self->get_left_sibling($edge);
+    if ( defined $left_sibling ) {
+        return ( $left_sibling->child->fields->[$field_index] );
+    } else {
+        return '#start#';
+    }
+}
+
+sub feature_right_sibling {
+    my ( $self, $edge, $field_index ) = @_;
+
+    my $right_sibling = $self->get_right_sibling($edge);
+    if ( defined $right_sibling ) {
+        return ( $right_sibling->child->fields->[$field_index] );
+    } else {
+        return '#end#';
+    }
+}
+
+sub get_left_sibling {
+    my ( $self, $edge ) = @_;
+
     my $siblings = $edge->parent->children;
     my $is_first = ( $siblings->[0]->child->ord == $edge->child->ord );
     if ($is_first) {
 
         # there is no left sibling to the leftmost node
-        return '#start#';
+        return;
     } else {
 
         # find my position among parent's children (is at least 1)
@@ -588,12 +796,12 @@ sub feature_left_sibling {
         }
 
         # now ($my_index-1) is the index of my (closest) left sibling
-        return ( $siblings->[ $my_index - 1 ]->child->fields->[$field_index] );
+        return ( $siblings->[ $my_index - 1 ] );
     }
 }
 
-sub feature_right_sibling {
-    my ( $self, $edge, $field_index ) = @_;
+sub get_right_sibling {
+    my ( $self, $edge ) = @_;
 
     my $siblings           = $edge->parent->children;
     my $last_sibling_index = scalar(@$siblings) - 1;
@@ -604,7 +812,7 @@ sub feature_right_sibling {
     if ($is_last) {
 
         # there is no right sibling to the rightmost node
-        return '#end#';
+        return;
     } else {
 
         # find my position among parent's children
@@ -615,7 +823,7 @@ sub feature_right_sibling {
         }
 
         # now ($my_index+1) is the index of my (closest) right sibling
-        return ( $siblings->[ $my_index + 1 ]->child->fields->[$field_index] );
+        return $siblings->[ $my_index + 1 ];
     }
 }
 
@@ -1206,6 +1414,134 @@ sub feature_number_of_parents_children {
     }
 }
 
+sub feature_additional_model {
+    my ( $self, $edge, $field_index, $model ) = @_;
+
+    my $child  = $edge->child->fields->[$field_index];
+    my $parent = $edge->parent->fields->[$field_index];
+
+    if ( defined $child && defined $parent ) {
+        return $model->get_value( $child, $parent );
+    } else {
+        croak "Either child or parent is undefined in additional model feature, " .
+            "this should not happen!";
+    }
+}
+
+sub feature_additional_model_bucketed {
+    my ( $self, $edge, $field_index, $model ) = @_;
+
+    my $child  = $edge->child->fields->[$field_index];
+    my $parent = $edge->parent->fields->[$field_index];
+
+    if ( defined $child && defined $parent ) {
+        return $model->get_bucketed_value( $child, $parent );
+    } else {
+        croak "Either child or parent is undefined in additional model feature, " .
+            "this should not happen!";
+    }
+}
+
+sub feature_additional_model_rounded {
+    my ( $self, $edge, $parameters, $model ) = @_;
+
+    my ( $field_index, $rounding ) = @$parameters;
+    my $child  = $edge->child->fields->[$field_index];
+    my $parent = $edge->parent->fields->[$field_index];
+
+    if ( defined $child && defined $parent ) {
+        return $model->get_rounded_value( $child, $parent, $rounding );
+    } else {
+        croak "Either child or parent is undefined in additional model feature, " .
+            "this should not happen!";
+    }
+}
+
+sub feature_additional_model_d {
+    my ( $self, $edge, $parameters, $model ) = @_;
+
+    my ( $field_index_c, $field_index_p ) = @$parameters;
+    my $child  = $edge->child->fields->[$field_index_c];
+    my $parent = $edge->parent->fields->[$field_index_p];
+
+    if ( defined $child && defined $parent ) {
+        return $model->get_rounded_value( $child, $parent );
+    } else {
+        croak "Either child or parent is undefined in additional model feature, " .
+            "this should not happen!";
+    }
+}
+
+sub feature_pmi {
+    my ( $self, $edge, $field_index ) = @_;
+
+    return $self->feature_additional_model( $edge, $field_index, $self->pmi_model );
+}
+
+sub feature_pmi_bucketed {
+    my ( $self, $edge, $field_index ) = @_;
+
+    return $self->feature_additional_model_bucketed( $edge, $field_index, $self->pmi_model );
+}
+
+sub feature_pmi_rounded {
+    my ( $self, $edge, $parameters ) = @_;
+
+    return $self->feature_additional_model_rounded( $edge, $parameters, $self->pmi_model );
+}
+
+sub feature_pmi_d {
+    my ( $self, $edge, $parameters ) = @_;
+
+    return $self->feature_additional_model_d( $edge, $parameters, $self->pmi_model );
+}
+
+sub feature_pmi_2_rounded {
+    my ( $self, $edge, $field_index ) = @_;
+
+    my @params = ( $field_index, 1 );
+    return $self->feature_pmi_rounded( $edge, \@params );
+}
+
+sub feature_pmi_3_rounded {
+    my ( $self, $edge, $field_index ) = @_;
+
+    my @params = ( $field_index, 2 );
+    return $self->feature_pmi_rounded( $edge, \@params );
+}
+
+sub feature_cprob {
+    my ( $self, $edge, $field_index ) = @_;
+
+    return $self->feature_additional_model( $edge, $field_index, $self->cprob_model );
+}
+
+sub feature_cprob_bucketed {
+    my ( $self, $edge, $field_index ) = @_;
+
+    return $self->feature_additional_model_bucketed( $edge, $field_index, $self->cprob_model );
+}
+
+sub feature_cprob_rounded {
+    my ( $self, $edge, $parameters ) = @_;
+
+    return $self->feature_additional_model_rounded( $edge, $parameters, $self->cprob_model );
+}
+
+sub feature_cprob_2_rounded {
+    my ( $self, $edge, $field_index ) = @_;
+
+    my @params = ( $field_index, 1 );
+    return $self->feature_cprob_rounded( $edge, \@params );
+}
+
+sub feature_cprob_3_rounded {
+    my ( $self, $edge, $field_index ) = @_;
+
+    my @params = ( $field_index, 2 );
+    return $self->feature_cprob_rounded( $edge, \@params );
+}
+
 1;
 
 __END__
@@ -1222,7 +1558,7 @@ Treex::Tool::Parser::MSTperl::FeaturesControl
 
 =head1 VERSION
 
-version 0.08268
+version 0.11306
 
 =head1 DESCRIPTION
 
